/*
 * pui.c
 *
 * Copyright (c) 1996, Scotter <scotter@iname.com>
 *
 * This is free software, licensed under the GNU Public License V2.
 * See the file LICENSE for details.
 *
 */

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "pui.h"
#include "pui_io.h"

char *pui_Name(PUI_FILE db)
{
  static char name[33]; /* Valid until next call */
  struct DBInfo info;
  pui_ReadDBInfo(db, &info);
  memcpy(name, info.name, 32);
  name[32] = 0; /* Force NULL termination safety */
  return name;
}

/* Returned data is valid until the next call */
char *pui_Version(PUI_FILE db)
{
  static char *v = 0; /* gdbm mallocs, we free */
  if (v) {
    free(v); /* Free old version result */
    v = 0;
  }
  pui_version_read(db, &v);
  return v;
}

unsigned long pui_First(PUI_FILE db)
{
  struct pui_global_record r;

  if (pui_global_read(db, &r) == 0) {
    return r.head;
  }
  return 0;
}

unsigned long pui_Last(PUI_FILE db)
{
  struct pui_global_record r;

  if (pui_global_read(db, &r) == 0) {
    return r.tail;
  }
  return 0;
}

unsigned long pui_Prev(PUI_FILE db, unsigned long id)
{
  struct pui_header_record r;

  if (pui_header_read(db, id, &r) == 0) {
    return r.prev;
  }
  return 0;
}

unsigned long pui_Next(PUI_FILE db, unsigned long id)
{
  struct pui_header_record r;

  if (pui_header_read(db, id, &r) == 0) {
    return r.next;
  }
  return 0;
}

unsigned long pui_NumRecords(PUI_FILE db)
{
  struct pui_global_record r;

  if (pui_global_read(db, &r) == 0) {
    return r.num_records;
  }
  return 0;
}

unsigned long pui_GetByIndex(PUI_FILE db, int i)
{
  unsigned long id;

  id = pui_First(db);
  while (id && i--) {
    id = pui_Next(db, id);
  }
  return id;
}

int pui_Exists(PUI_FILE db, unsigned long id)
{
  return (pui_header_exists(db, id) & pui_record_exists(db, id));
}

unsigned long pui_UniqueId(PUI_FILE db)
{
  unsigned long id = 1;

  while(id && pui_Exists(db, id)) ++id; /* End if loop to zero */

  return id;
}

void pui_ReadHeader(PUI_FILE db, unsigned long id, int * attr, int * category)
{
  struct pui_header_record header;
  assert(pui_header_exists(db, id));
  pui_header_read(db, id, &header);
  *attr = header.attr;
  *category = header.category;
}

void pui_WriteHeader(PUI_FILE db, unsigned long id, int attr, int category)
{
  struct pui_header_record header;
  assert(pui_header_exists(db, id));
  pui_header_read(db, id, &header); /* Preserve other attributes of header */
  header.attr = attr;
  header.category = category;
  pui_header_write(db, id, &header);
}

void pui_ReadRecord(PUI_FILE db, unsigned long id, int * size, char * buffer)
{
  struct pui_data_record data;
  assert(pui_record_exists(db, id));
  data.data = buffer; /* Copy directly into provided buffer */
  pui_record_read(db, id, &data);
  *size = data.length;
}

void pui_WriteRecord(PUI_FILE db, unsigned long id, int size, char * buffer)
{
  struct pui_data_record data;
  assert(pui_record_exists(db, id));
  data.length = size;
  data.data = buffer; /* Copy directly from provided buffer */
  pui_record_write(db, id, &data); /* Trashes old record completely */
}

void pui_Read(PUI_FILE db, unsigned long id, int *attr, int *category, int *size, char *buffer)
{
  pui_ReadHeader(db, id, attr, category);
  pui_ReadRecord(db, id, size, buffer);
}

void pui_Write(PUI_FILE db, unsigned long id, int attr, int category, int size, char * buffer)
{
  pui_WriteHeader(db, id, attr, category);
  pui_WriteRecord(db, id, size, buffer);
}

void pui_ReadDBInfo(PUI_FILE db, struct DBInfo *info)
{
  pui_dbinfo_read(db, info);
}

void pui_WriteDBInfo(PUI_FILE db, struct DBInfo *info)
{
  pui_dbinfo_write(db, info);
}

unsigned long pui_Insert(
  PUI_FILE db,
  unsigned long nextid,
  unsigned long id,
  int attr,
  int category,
  int size,
  char *buffer)
{
  struct pui_global_record global;
  struct pui_header_record header;
  struct pui_header_record h;
  struct pui_data_record d;

  h.lock = 0;
  h.attr = attr;
  h.category = category;

  if (id == 0) {
    id = pui_UniqueId(db);
  }
  /* Else should check to see if id exists! */

  pui_global_read(db, &global);

  h.next = nextid;
  if (nextid) {
    pui_header_read(db, nextid, &header);
    h.prev = header.prev;
    header.prev = id;
    pui_header_write(db, nextid, &header);
  } else {
    h.prev = global.tail;
    global.tail = id;
  }
    
  if (h.prev) {
    pui_header_read(db, h.prev, &header);
    header.next = id;
    pui_header_write(db, h.prev, &header);
  } else {
    global.head = id;
  }

  pui_header_write(db, id, &h);

  d.length = size;
  d.data = buffer;
  pui_record_write(db, id, &d);

  global.num_records++;
  pui_global_write(db, &global);

  return id;
}

unsigned long pui_InsertAtIndex(
  PUI_FILE db,
  int i,
  unsigned long id,
  int attr,
  int category,
  int size,
  char *buffer)
{
  unsigned long nextid;
  struct pui_global_record global;
  struct pui_header_record header;

  pui_global_read(db, &global);

  if (i < 0 || i >= global.num_records) /* Append to list */
    nextid = 0;
  else
    nextid = pui_GetByIndex(db, i);

  return pui_Insert(db, nextid, id, attr, category, size, buffer);
}

void pui_Delete(PUI_FILE db, unsigned long id)
{
  struct pui_global_record global;
  struct pui_header_record header;
  struct pui_header_record tmpheader;

  pui_global_read(db, &global);

  pui_header_read(db, id, &header);
  if (header.prev) {
    pui_header_read(db, header.prev, &tmpheader);
    tmpheader.next = header.next;
    pui_header_write(db, header.prev, &tmpheader);
  } else {
    global.head = header.next;
  }
  if (header.next) {
    pui_header_read(db, header.next, &tmpheader);
    tmpheader.prev = header.prev;
    pui_header_write(db, header.next, &tmpheader);
  } else {
    global.tail = header.prev;
  }

  pui_header_delete(db, id);
  pui_record_delete(db, id);

  global.num_records--;

  pui_global_write(db, &global);
}

void pui_DeleteByIndex(PUI_FILE db, int i)
{
  unsigned long id = pui_GetByIndex(db, i);
  pui_Delete(db, id);
}

void pui_ReadAppInfo(PUI_FILE db, int *size, char *buffer)
{
  struct pui_data_record data;
  
  data.data = buffer; /* Require user to pass in storage space */
  pui_appinfo_read(db, &data);

  *size = data.length;
}

void pui_WriteAppInfo(PUI_FILE db, int size, char *buffer)
{
  struct pui_data_record data;
  data.length = size;
  data.data = buffer; /* Copy directly from provided buffer */
  pui_appinfo_write(db, &data); /* Trashes old record */
}
