/*
 * pui_io.c
 *
 * Copyright (c) 1996, Scotter <scotter@iname.com>
 *
 * This is free software, licensed under the GNU Public License V2.
 * See the file LICENSE for details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <gdbm.h>

#include "pui.h"
#include "pui_io.h"

/*
 * This file contains ALL gdbm file access routines.
 * It is the ONLY file with gdbm.h.
 * To best isolate the database implementation, all other
 * routines should call those in here.
 *
 * Also, for the most part, these routines will not be public
 * outside of this library, thus they are not in pui.h
 *
 * Data reading and writing is performed in an architecture
 * independent manner, using the following set of macros.
 * These match the ones in pi-socket.h, and should probably
 * be moved to a higher level "pilot.h"
 *
 * Routines that deal with variable length structures can
 * use buffers passed in by the caller, or will malloc
 * and pass back pointers. The caller should be aware
 * of how they are calling the routine and take care
 * of managing the memory.
 */

/* Use with unsigned char ptr's! */
#define get_long(ptr) (((ptr)[0] << 24) | \
                       ((ptr)[1] << 16) | \
                       ((ptr)[2] << 8)  | \
                       ((ptr)[3] << 0))

#define get_short(ptr) (((ptr)[0] << 8)  | \
                        ((ptr)[1] << 0))

#define get_byte(ptr) ((ptr)[0])

#define set_long(ptr,val) (((ptr)[0] = ((val) >> 24) & 0xff), \
                          ((ptr)[1] = ((val) >> 16) & 0xff), \
                          ((ptr)[2] = ((val) >> 8) & 0xff), \
                          ((ptr)[3] = ((val) >> 0) & 0xff))

#define set_short(ptr,val) (((ptr)[0] = ((val) >> 8) & 0xff), \
                            ((ptr)[1] = ((val) >> 0) & 0xff))

#define set_byte(ptr,val) ((ptr)[0]=(val))

/* These are "special" one-byte keys for storing global information */
/* in the database. */

#define PUI_VERSION_KEY 0
#define PUI_DBINFO_KEY 1
#define PUI_DBINFO_SIZE 60
#define PUI_GLOBAL_KEY 17
#define PUI_GLOBAL_SIZE 16
#define PUI_HEADER_SIZE 14
#define PUI_APPINFO_KEY 21

/* Adding APPINFO did not require bumping up the version. */
/* Changes since gdbm file version 0.0.1: */
/*   The "name" token is now complete DBInfo data */
static char *fileversion = "0.0.2";

/* This is a magic number, used to pad an id to make a header key. */
#define PUI_HEADER_MAGIC 42

/*
 * NEW/OPEN/CLOSE
 *
 */
PUI_FILE pui_Create(char *filename, struct DBInfo *info)
{
  struct pui_global_record r;
  PUI_FILE db;

  r.lock = 0;
  r.num_records = 0;
  r.head = 0;
  r.tail = 0;
  db = (PUI_FILE)gdbm_open(filename, 0, GDBM_NEWDB, 0666, 0);
  if (db) {
    pui_global_write(db, &r);
    if (info) {
      pui_dbinfo_write(db, info);
    }
    pui_version_write(db, &fileversion);
  }
  return db;
}

PUI_FILE pui_Open(char *name, const char *mode)
{
  int md;
  PUI_FILE db;
  char *version = 0;

  md = GDBM_READER;
  if (mode && (mode[0] == 'w')) md = GDBM_WRITER;

  db = (PUI_FILE)gdbm_open(name, 0, md, 0666, 0);
  if (db) {
    pui_version_read(db, &version);
    if (version == 0 || strcmp(version, fileversion) != 0) {
      fprintf(stderr, "File version mismatch: file is %s, libpui is %s\n", version, fileversion);
      gdbm_close((GDBM_FILE)db);
      db = 0;
    }
    if (version) free(version);
  }

  return db;
}

void pui_Close(PUI_FILE db)
{
  if (!pui_dbinfo_exists(db)) {
    fprintf(stderr, "Warning: pui database has no dbinfo record\n");
  }
  gdbm_close((GDBM_FILE)db);
}

void pui_Reorganize(PUI_FILE db)
{
  gdbm_reorganize((GDBM_FILE)db);
}

/*
 * VERSION
 *
 */
int pui_version_read(
  const PUI_FILE db,
  char **buf) /* double-star so user can provide buffer or receive fetch */
{
  datum key;
  unsigned char keydata[1];
  datum data;

  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_VERSION_KEY);

  data = gdbm_fetch((GDBM_FILE)db, key);

  if (!data.dptr) return -1;

  if (*buf) { /* If a buffer pointer was passed in, use it. */
    memcpy(*buf, data.dptr, data.dsize);
    free(data.dptr);
  } else {    /* Otherwise, return the space gdbm malloc'd. User must free. */
    *buf = data.dptr;
  }

  return 0;
}

int pui_version_write(
  const PUI_FILE db,
  char **buf)  /* double-star for consistency */
{
  datum key;
  unsigned char keydata[1];
  datum data;
  
  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_VERSION_KEY);

  data.dsize = strlen(*buf) + 1; /* Include terminating NULL */
  data.dptr = *buf; /* This prevents **buf from being const */

  return gdbm_store((GDBM_FILE)db, key, data, GDBM_REPLACE);
}

int pui_version_exists(const PUI_FILE db)
{
  datum key;
  unsigned char keydata[1];
  
  key.dsize = 1;
  key.dptr = keydata;
  set_byte(keydata, PUI_VERSION_KEY);

  return gdbm_exists((GDBM_FILE)db, key);
}

int pui_version_delete(const PUI_FILE db)
{
  datum key;
  unsigned char keydata[1];

  key.dsize = 1;
  key.dptr = keydata;
  set_byte(keydata, PUI_VERSION_KEY);

  return gdbm_delete((GDBM_FILE)db, key);
}

#define PILOT_TIME_DELTA (unsigned)(2082844800)
#define pilot_time_to_unix_time(t) ((time_t)((t) - PILOT_TIME_DELTA))
#define unix_time_to_pilot_time(t) ((unsigned long)((t) + PILOT_TIME_DELTA))

/*
 * DBINFO
 *
 */
int pui_dbinfo_read(
  const PUI_FILE db,
  struct DBInfo *ip)
{
  datum key;
  unsigned char keydata[1];
  datum data;
  unsigned char *dptr;

  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_DBINFO_KEY);

  data = gdbm_fetch((GDBM_FILE)db, key);

  if (!data.dptr) return -1;

  dptr = (unsigned char *)data.dptr;

  assert(data.dsize == PUI_DBINFO_SIZE);

  /* This is VERY similar to pi-file reading. */
  memcpy(ip->name, dptr, 32);
  ip->flags       = get_short(dptr+32);
  ip->version     = get_short(dptr+34);
  ip->crdate      = pilot_time_to_unix_time (get_long(dptr+36));
  ip->moddate     = pilot_time_to_unix_time (get_long(dptr+40));
  ip->backupdate  = pilot_time_to_unix_time (get_long(dptr+44));
  ip->modnum      = get_long(dptr+48);
  ip->type        = get_long(dptr+52);
  ip->creator     = get_long(dptr+56);
  ip->more  = 0; /* Not stored */
  ip->index = 0; /* Not stored */

  free(data.dptr);

  return 0;
}

int pui_dbinfo_write(
  const PUI_FILE db,
  const struct DBInfo *ip)
{
  datum key;
  unsigned char keydata[1];
  datum data;
  unsigned char datadata[PUI_DBINFO_SIZE];
  
  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_DBINFO_KEY);

  data.dsize = PUI_DBINFO_SIZE;
  data.dptr = datadata;
  memcpy(datadata, ip->name, 32);
  set_short(datadata+32, ip->flags);
  set_short(datadata+34, ip->version);
  set_long(datadata+36, unix_time_to_pilot_time(ip->crdate));
  set_long(datadata+40, unix_time_to_pilot_time(ip->moddate));
  set_long(datadata+44, unix_time_to_pilot_time(ip->backupdate));
  set_long(datadata+48, ip->modnum);
  set_long(datadata+52, ip->type);
  set_long(datadata+56, ip->creator);

  return gdbm_store((GDBM_FILE)db, key, data, GDBM_REPLACE);
}

int pui_dbinfo_exists(const PUI_FILE db)
{
  datum key;
  unsigned char keydata[1];
  
  key.dsize = 1;
  key.dptr = keydata;
  set_byte(keydata, PUI_DBINFO_KEY);

  return gdbm_exists((GDBM_FILE)db, key);
}

int pui_dbinfo_delete(const PUI_FILE db)
{
  datum key;
  unsigned char keydata[1];

  key.dsize = 1;
  key.dptr = keydata;
  set_byte(keydata, PUI_DBINFO_KEY);

  return gdbm_delete((GDBM_FILE)db, key);
}

/*
 * GLOBAL
 *
 */
int pui_global_read(
  const PUI_FILE db,
  struct pui_global_record *r)
{
  datum key;
  unsigned char keydata[1];
  datum data;
  unsigned char *dptr;

  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_GLOBAL_KEY);

  data = gdbm_fetch((GDBM_FILE)db, key);

  if (!data.dptr) return -1;

  dptr = (unsigned char *)data.dptr; /* Force unsigned */

  assert(data.dsize == PUI_GLOBAL_SIZE);

  r->lock        = get_long(dptr);
  r->num_records = get_long(dptr+4);
  r->head        = get_long(dptr+8);
  r->tail        = get_long(dptr+12);

  free(data.dptr);
  return 0;
}

int pui_global_write(
  const PUI_FILE db,
  const struct pui_global_record *r)
{
  datum key;
  unsigned char keydata[1];
  datum data;
  unsigned char datadata[PUI_GLOBAL_SIZE];
  
  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_GLOBAL_KEY);

  data.dsize = PUI_GLOBAL_SIZE;
  data.dptr = datadata;
  set_long(datadata   , r->lock);
  set_long(datadata+4 , r->num_records);
  set_long(datadata+8 , r->head);
  set_long(datadata+12, r->tail);

  return gdbm_store( (GDBM_FILE)db, key, data, GDBM_REPLACE);
}

int pui_global_exists(const PUI_FILE db)
{
  datum key;
  unsigned char keydata[1];
  
  key.dsize = 1;
  key.dptr = keydata;
  set_byte(keydata, PUI_GLOBAL_KEY);

  return gdbm_exists((GDBM_FILE)db, key);
}

int pui_global_delete(const PUI_FILE db)
{
  datum key;
  unsigned char keydata[1];

  key.dsize = 1;
  key.dptr = keydata;
  set_byte(keydata, PUI_GLOBAL_KEY);

  return gdbm_delete((GDBM_FILE)db, key);
}

/*
 * HEADER
 *
 */
int pui_header_read(
  const PUI_FILE db,
  const unsigned long id,
  struct pui_header_record *r)
{
  datum key;
  unsigned char keydata[5];
  datum data;
  unsigned char *dptr;
  
  key.dsize = 5;
  key.dptr = keydata;
  set_long(keydata, id);
  set_byte(keydata+4, PUI_HEADER_MAGIC);

  data = gdbm_fetch((GDBM_FILE)db, key);

  if (!data.dptr) return -1;

  dptr = (unsigned char *)data.dptr;

  assert(data.dsize == PUI_HEADER_SIZE);

  r->lock     = get_long(dptr);
  r->attr     = get_byte(dptr+4);
  r->category = get_byte(dptr+5);
  r->prev     = get_long(dptr+6);
  r->next     = get_long(dptr+10);

  free(data.dptr);

  return 0;
}

int pui_header_write(
  const PUI_FILE db,
  const unsigned long id,
  const struct pui_header_record *r)
{
  datum key;
  unsigned char keydata[5];
  datum data;
  unsigned char datadata[PUI_HEADER_SIZE];
  
  key.dsize = 5;
  key.dptr = keydata;
  set_long(keydata, id);
  set_byte(keydata+4, PUI_HEADER_MAGIC);

  data.dsize = PUI_HEADER_SIZE;
  data.dptr = datadata;
  set_long(datadata,    r->lock);
  set_byte(datadata+4,  r->attr);
  set_byte(datadata+5,  r->category);
  set_long(datadata+6,  r->prev);
  set_long(datadata+10, r->next);

  return gdbm_store( (GDBM_FILE)db, key, data, GDBM_REPLACE);
}

int pui_header_exists(const PUI_FILE db, const unsigned long id)
{
  datum key;
  unsigned char keydata[5];
  
  key.dsize = 5;
  key.dptr = keydata;
  set_long(keydata, id);
  set_byte(keydata+4, PUI_HEADER_MAGIC);

  return gdbm_exists((GDBM_FILE)db, key);
}

int pui_header_delete(const PUI_FILE db, const unsigned long id)
{
  datum key;
  unsigned char keydata[5];

  key.dsize = 5;
  key.dptr = keydata;
  set_long(keydata, id);
  set_byte(keydata+4, PUI_HEADER_MAGIC);

  return gdbm_delete((GDBM_FILE)db, key);
}

/*
 * RECORD
 *
 */
int pui_record_read(
  const PUI_FILE db,
  const unsigned long id,
  struct pui_data_record *r)
{
  datum key;
  unsigned char keydata[4];
  datum data;
  
  key.dsize = 4; /* sizeof(long) */
  key.dptr = keydata;
  set_long(keydata, id);

  data = gdbm_fetch((GDBM_FILE)db, key);

  if (!data.dptr) {
    r->length = 0;
    return -1;
  }

  r->length = data.dsize;
  if (r->data) { /* If user provides data buffer, use it. */
    memcpy(r->data, data.dptr, data.dsize);
    free(data.dptr);
  } else {       /* Othersize return data gdbm malloc'd. User must free. */
    r->data = data.dptr;
  }

  return 0;
}

int pui_record_write(
  const PUI_FILE db,
  const unsigned long id,
  const struct pui_data_record *r)
{
  datum key;
  unsigned char keydata[4];
  datum data;
  
  key.dsize = 4;
  key.dptr = keydata;
  set_long(keydata, id);

  data.dsize = r->length;
  data.dptr = r->data;

  return gdbm_store( (GDBM_FILE)db, key, data, GDBM_REPLACE);
}

int pui_record_exists(const PUI_FILE db, const unsigned long id)
{
  datum key;
  unsigned char keydata[4];
  
  key.dsize = 4;
  key.dptr = keydata;
  set_long(keydata, id);

  return gdbm_exists((GDBM_FILE)db, key);
}

int pui_record_delete(const PUI_FILE db, const unsigned long id)
{
  datum key;
  unsigned char keydata[4];

  key.dsize = 4;
  key.dptr = keydata;
  set_long(keydata, id);

  return gdbm_delete((GDBM_FILE)db, key);
}

int pui_appinfo_read(const PUI_FILE db, struct pui_data_record *r)
{
  datum key;
  unsigned char keydata[1];
  datum data;

  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_APPINFO_KEY);

  data = gdbm_fetch((GDBM_FILE)db, key);

  if (!data.dptr) {
    r->length = 0;
    return 0;
  }

  r->length = data.dsize;
  if (r->data) { /* If user provides data buffer, use it. */
    memcpy(r->data, data.dptr, data.dsize);
    free(data.dptr);
  } else {       /* Othersize return data gdbm malloc'd. User must free. */
    r->data = data.dptr;
  }

  return 0; 
}

int pui_appinfo_write(const PUI_FILE db, const struct pui_data_record *r)
{
  datum key;
  unsigned char keydata[1];
  datum data;

  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_APPINFO_KEY);

  data.dsize = r->length;
  data.dptr = r->data;

  return gdbm_store( (GDBM_FILE)db, key, data, GDBM_REPLACE);
}

int pui_appinfo_exists(const PUI_FILE db)
{
  datum key;
  unsigned char keydata[1];
  
  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_APPINFO_KEY);

  return gdbm_exists((GDBM_FILE)db, key);
}

int pui_appinfo_delete(const PUI_FILE db)
{
  datum key;
  unsigned char keydata[1];
  
  key.dsize = 1; /* magic number only */
  key.dptr = keydata;
  set_byte(keydata, PUI_APPINFO_KEY);

  return gdbm_delete((GDBM_FILE)db, key);
}
