/*
 * tclAddressDB.c
 *
 * Copyright (c) 1996, Scotter <scotter@iname.com>
 *
 * This is free software, licensed under the GNU Public License V2.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tcl.h"
#include "pi-address.h"
#include "pui.h"
#include "tclpui.h"

static void tclAddressDB_NewDBInfo(struct DBInfo *);
static void tclAddressDB_Read(Tcl_Interp *, char *, int , char *);
static void tclAddressDB_Write(Tcl_Interp *, char *, int *, char *);
static void tclAddressDB_AppInfoRead(Tcl_Interp *, char *, PUI_FILE);
static void tclAddressDB_AppInfoWrite(Tcl_Interp *, char *, PUI_FILE);

/*
 * Define the public symbol "object" through which we access the routines
 * in this module.
 */
struct tclDBModule tclAddressDB = {
  "AddressDB",  tclAddressDB_NewDBInfo,
                tclAddressDB_Read,
                tclAddressDB_Write,
                tclAddressDB_AppInfoRead,
                tclAddressDB_AppInfoWrite };

/*---------------------AddressDB-----------------------*/

/* These match enum in address.h */
static char *entrynames[] = {
  "lastname", "firstname", "company",
  "phone1", "phone2", "phone3", "phone4", "phone5",
  "address", "city", "state", "zip", "country", "title",
  "custom1", "custom2", "custom3", "custom4",
  "note"
};

static void tclAddressDB_NewDBInfo(struct DBInfo *info)
{
  info->type = 0x44415441;
  info->creator = 0x61646472;
  info->crdate = 818989054;
  info->moddate = 818989054;
  info->backupdate = -2082816000;
  info->modnum = 1;
  info->flags = 0;
  info->version = 0;
  strncpy(info->name, "AddressDB", 32);
  info->index = 0;
  info->more = 0;
}

static void tclAddressDB_Read(Tcl_Interp *interp, char *varname, int size, char *buffer)
{
  int i;
  char s[16];
  struct Address addr;
  unpack_Address(&addr, buffer, size);

  for (i = 0; i < 19; ++i) {
    Tcl_SetVar2(interp, varname, entrynames[i], SAFE(addr.entry[i]), 0);
  }

  sprintf(s, "%d", (int)addr.phonelabel[0]);
  Tcl_SetVar2(interp, varname, "phonelabel1", s, 0);
  sprintf(s, "%d", (int)addr.phonelabel[1]);
  Tcl_SetVar2(interp, varname, "phonelabel2", s, 0);
  sprintf(s, "%d", (int)addr.phonelabel[2]);
  Tcl_SetVar2(interp, varname, "phonelabel3", s, 0);
  sprintf(s, "%d", (int)addr.phonelabel[3]);
  Tcl_SetVar2(interp, varname, "phonelabel4", s, 0);
  sprintf(s, "%d", (int)addr.phonelabel[4]);
  Tcl_SetVar2(interp, varname, "phonelabel5", s, 0);
  sprintf(s, "%d", (int)addr.whichphone);
  Tcl_SetVar2(interp, varname, "whichphone", s, 0);
  free_Address(&addr);
}

static void tclAddressDB_Write(Tcl_Interp *interp, char *varname, int *size, char *buffer)
{
  struct Address addr;
  char *value;
  int i;

  for (i = 0; i < 19; ++i) {
    value = Tcl_GetVar2(interp, varname, entrynames[i], 0);
    addr.entry[i] = value?(*value?strdup(value):0):0;
  }

  value = Tcl_GetVar2(interp, varname, "phonelabel1", 0);
  addr.phonelabel[0] = value?atoi(value):0;
  value = Tcl_GetVar2(interp, varname, "phonelabel2", 0);
  addr.phonelabel[1] = value?atoi(value):1;
  value = Tcl_GetVar2(interp, varname, "phonelabel3", 0);
  addr.phonelabel[2] = value?atoi(value):2;
  value = Tcl_GetVar2(interp, varname, "phonelabel4", 0);
  addr.phonelabel[3] = value?atoi(value):3;
  value = Tcl_GetVar2(interp, varname, "phonelabel5", 0);
  addr.phonelabel[4] = value?atoi(value):4;

  value = Tcl_GetVar2(interp, varname, "whichphone", 0);
  addr.whichphone = value?atoi(value):1;
  if (addr.entry[entryPhone1 + addr.whichphone] == 0) {
    addr.whichphone = 0;
    for (i = 0; i < 5; ++i) { /* Find an existing phone */
      if (addr.entry[entryPhone1 + i]) {
        addr.whichphone = i;
        break; /* out of for */
      }
    }
  }

  pack_Address(&addr, buffer, size);
  free_Address(&addr);
}

static void tclAddressDB_AppInfoRead(Tcl_Interp *interp, char *varname, PUI_FILE puidb)
{
  char s[32];
  struct AddressAppInfo appinfo;
  int i;
  char buffer[0xffff];
  int size;

  pui_ReadAppInfo(puidb, &size, buffer);
  unpack_AddressAppInfo(&appinfo, buffer, size);

  sprintf(s, "%u", appinfo.renamedcategories);
  Tcl_SetVar2(interp, varname, "renamedcategories", s, 0);

  for (i=0;i<16;++i) {
    sprintf(s, "CategoryName,%d", i);
    Tcl_SetVar2(interp, varname, s, appinfo.CategoryName[i], 0);
  }
  Tcl_SetVar2(interp, varname, "CategoryID", appinfo.CategoryID, 0);

  sprintf(s, "%u", appinfo.lastUniqueID);
  Tcl_SetVar2(interp, varname, "lastUniqueID", s, 0);

  sprintf(s, "%d", appinfo.dirtyfieldlabels);
  Tcl_SetVar2(interp, varname, "dirtyfieldlabels", s, 0);

  for (i=0;i<19+3;++i) {
    sprintf(s, "labels,%d", i);
    Tcl_SetVar2(interp, varname, s, appinfo.labels[i], 0);
  }
  /* Should we output the duplicates??? */
  for (i=0;i<8;++i) {
    sprintf(s, "phonelabels,%d", i);
    Tcl_SetVar2(interp, varname, s, appinfo.phonelabels[i], 0);
  }
  sprintf(s, "%d", appinfo.country);
  Tcl_SetVar2(interp, varname, "country", s, 0);
  sprintf(s, "%d", appinfo.sortByCompany);
  Tcl_SetVar2(interp, varname, "sortByCompany", s, 0);
}

static void tclAddressDB_AppInfoWrite(Tcl_Interp *interp, char *varname, PUI_FILE puidb)
{
  char s[32];
  struct AddressAppInfo appinfo;
  char *value;
  int i;
  char buffer[0xffff];
  int size;

  value = Tcl_GetVar2(interp, varname, "renamedcategories", 0);
  appinfo.renamedcategories = SAFESTRTOUL(value,10);

  for (i=0;i<16;++i) {
    sprintf(s, "CategoryName,%d", i);
    value = Tcl_GetVar2(interp, varname, s, 0);
    SAFESTRINGCOPY(appinfo.CategoryName[i],value);
  }

  value = Tcl_GetVar2(interp, varname, "CategoryID", 0);
  SAFESTRINGCOPY(appinfo.CategoryID,value);

  value = Tcl_GetVar2(interp, varname, "lastUniqueID", 0);
  appinfo.lastUniqueID = (unsigned char)SAFESTRTOUL(value,10);

  value = Tcl_GetVar2(interp, varname, "dirtyfieldlabels", 0);
  appinfo.dirtyfieldlabels = SAFESTRTOUL(value, 10);

  for (i=0;i<19+3;++i) {
    sprintf(s, "labels,%d", i);
    value = Tcl_GetVar2(interp, varname, s, 0);
    SAFESTRINGCOPY(appinfo.labels[i],value);
  }
  for (i=0;i<8;++i) {
    sprintf(s, "phonelabels,%d", i);
    value = Tcl_GetVar2(interp, varname, s, 0);
    SAFESTRINGCOPY(appinfo.phonelabels[i],value);
  }
  value = Tcl_GetVar2(interp, varname, "country", 0);
  appinfo.country = SAFEATOI(value);
  value = Tcl_GetVar2(interp, varname, "sortByCompany", 0);
  appinfo.sortByCompany = SAFEATOI(value);

  fprintf(stderr, "Warning, pack_AddressAppInfo is unimplemented.\n");
/*
 !!!! UNIMPLEMENTED
  pack_AddressAppInfo(&appinfo, buffer, &size);
*/
  pui_WriteAppInfo(puidb, size, buffer);
}
