/*
 * tclToDoDB.c
 *
 * Copyright (c) 1996, Scotter <scotter@iname.com>
 *
 * This is free software, licensed under the GNU Public License V2.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tcl.h"
#include "pi-todo.h"
#include "pui.h"
#include "tclpui.h"

static void tclToDoDB_NewDBInfo(struct DBInfo *);
static void tclToDoDB_Read(Tcl_Interp *, char *, int , char *);
static void tclToDoDB_Write(Tcl_Interp *, char *, int *, char *);
static void tclToDoDB_AppInfoRead(Tcl_Interp *, char *, PUI_FILE);
static void tclToDoDB_AppInfoWrite(Tcl_Interp *, char *, PUI_FILE);

/*
 * Define the public symbol "object" through which we access the routines
 * in this module.
 */
struct tclDBModule tclToDoDB = {
  "ToDoDB", tclToDoDB_NewDBInfo,
            tclToDoDB_Read,
            tclToDoDB_Write,
            tclToDoDB_AppInfoRead,
            tclToDoDB_AppInfoWrite
};

/*---------------------ToDoDB-----------------------*/

static void tclToDoDB_NewDBInfo(struct DBInfo *info)
{
  info->type = 0x44415441;
  info->creator = 0x746f646f;
  info->crdate = 818137282;
  info->moddate = 818137282;
  info->backupdate = -2082816000;
  info->modnum = 1;
  info->flags = 0;
  info->version = 0;
  strncpy(info->name, "ToDoDB", 32);
  info->index = 0;
  info->more = 0;
}

static void tclToDoDB_Read(Tcl_Interp *interp, char *varname, int size, char *buffer)
{
  char s[16];
  struct ToDo todo;
  time_t t;

  unpack_ToDo(&todo, buffer, size);

  sprintf(s, "%d", todo.indefinite);
  Tcl_SetVar2(interp, varname, "indefinite", s, 0);

  /* Convert for use with tcl "clock" format */
  if (todo.indefinite) {
    Tcl_SetVar2(interp, varname, "due", "", 0); /* Make a blank entry anyway */
  } else {
    t = mktime(&todo.due);
    sprintf(s, "%ld", (long)t);
    Tcl_SetVar2(interp, varname, "due", s, 0);
  }

  sprintf(s, "%d", todo.priority);
  Tcl_SetVar2(interp, varname, "priority", s, 0);
  sprintf(s, "%d", todo.complete);
  Tcl_SetVar2(interp, varname, "complete", s, 0);

  Tcl_SetVar2(interp, varname, "description", SAFE(todo.description), 0);
  Tcl_SetVar2(interp, varname, "note", SAFE(todo.note), 0);

  free_ToDo(&todo);
}

static void tclToDoDB_Write(Tcl_Interp *interp, char *varname, int *size, char *buffer)
{
  struct ToDo todo;
  char *value;
  time_t secs;
  struct tm *t;

  value = Tcl_GetVar2(interp, varname, "indefinite", 0);
  todo.indefinite = SAFEATOI(value);

  /* Convert for use with tcl "clock" format */
  if (todo.indefinite == 0) {
    value = Tcl_GetVar2(interp, varname, "due", 0);
    secs = (time_t)SAFESTRTOUL(value, 10);
    t = localtime(&secs);
    memcpy(&todo.due, t, sizeof(struct tm));
  }

  value = Tcl_GetVar2(interp, varname, "priority", 0);
  todo.priority = SAFEATOI(value);
  value = Tcl_GetVar2(interp, varname, "complete", 0);
  todo.complete = SAFEATOI(value);

  value = Tcl_GetVar2(interp, varname, "description", 0);
  todo.description = value?(*value?strdup(value):0):0;
  value = Tcl_GetVar2(interp, varname, "note", 0);
  todo.note = value?(*value?strdup(value):0):0;

  pack_ToDo(&todo, buffer, size);

  free_ToDo(&todo);
}

static void tclToDoDB_AppInfoRead(Tcl_Interp *interp, char *varname, PUI_FILE puidb)
{
  char s[32];
  struct ToDoAppInfo appinfo;
  int i;
  char buffer[0xffff];
  int size;

  pui_ReadAppInfo(puidb, &size, buffer);

  unpack_ToDoAppInfo(&appinfo, buffer, size);

  sprintf(s, "%u", appinfo.renamedcategories);
  Tcl_SetVar2(interp, varname, "renamedcategories", s, 0);

  for (i=0;i<16;++i) {
    sprintf(s, "CategoryName,%d", i);
    Tcl_SetVar2(interp, varname, s, appinfo.CategoryName[i], 0);
  }
  Tcl_SetVar2(interp, varname, "CategoryID", appinfo.CategoryID, 0);

  sprintf(s, "%u", appinfo.lastUniqueID);
  Tcl_SetVar2(interp, varname, "lastUniqueID", s, 0);

  sprintf(s, "%d", appinfo.dirty);
  Tcl_SetVar2(interp, varname, "dirty", s, 0);
  sprintf(s, "%d", appinfo.sortByPriority);
  Tcl_SetVar2(interp, varname, "sortByPriority", s, 0);
}

static void tclToDoDB_AppInfoWrite(Tcl_Interp *interp, char *varname, PUI_FILE puidb)
{
  char s[32];
  struct ToDoAppInfo appinfo;
  char *value;
  int i;
  char buffer[0xffff];
  int size;

  value = Tcl_GetVar2(interp, varname, "renamedcategories", 0);
  appinfo.renamedcategories = SAFESTRTOUL(value,10);

  for (i=0;i<16;++i) {
    sprintf(s, "CategoryName,%d", i);
    value = Tcl_GetVar2(interp, varname, s, 0);
    SAFESTRINGCOPY(appinfo.CategoryName[i],value);
  }

  value = Tcl_GetVar2(interp, varname, "CategoryID", 0);
  SAFESTRINGCOPY(appinfo.CategoryID,value);

  value = Tcl_GetVar2(interp, varname, "lastUniqueID", 0);
  appinfo.lastUniqueID = (unsigned char)SAFESTRTOUL(value,10);

  value = Tcl_GetVar2(interp, varname, "dirty", 0);
  appinfo.dirty = SAFEATOI(value);
  value = Tcl_GetVar2(interp, varname, "sortByPriority", 0);
  appinfo.sortByPriority = SAFEATOI(value);
  
  pack_ToDoAppInfo(&appinfo, buffer, &size);
  pui_WriteAppInfo(puidb, size, buffer);
}
