/*
 * tclpui.c
 *
 * Copyright (c) 1996, Scotter <scotter@iname.com>
 *
 * This is free software, licensed under the GNU Public License V2.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tcl.h"
#include "pi-dlp.h" /* For DBInfo structure */
#include "pui.h"
#include "tclpui.h"

static Tcl_CmdProc puiNew;
static Tcl_CmdProc puiOpen;
static Tcl_CmdProc puiClose;

static Tcl_CmdProc puiName;
static Tcl_CmdProc puiVersion;

static Tcl_CmdProc puiFirst;
static Tcl_CmdProc puiLast;
static Tcl_CmdProc puiPrev;
static Tcl_CmdProc puiNext;
static Tcl_CmdProc puiList;

static Tcl_CmdProc puiRead;
static Tcl_CmdProc puiWrite;
static Tcl_CmdProc puiDelete;

extern struct tclDBModule tclAddressDB;
extern struct tclDBModule tclDatebookDB;
extern struct tclDBModule tclMemoDB;
extern struct tclDBModule tclToDoDB;

/* These modules could be compiled as separate DSO's and loaded
   at runtime, but I'm hardwiring them here. */

struct tclDBModule *tclDBList[] = {
  &tclAddressDB,
  &tclDatebookDB,
  &tclMemoDB,
  &tclToDoDB,
  0 /* Terminate the list */
};

/* Search through tclDBList looking for given database. */
static int dbentry(char *name)
{
  int i = 0;

  while (tclDBList[i]) {
    if (strcmp(tclDBList[i]->name, name) == 0) return i;
    ++i;
  }
  return -1;
}


int Tclpui_Init(Tcl_Interp *i)
{
  Tcl_CreateCommand(i, "puiNew",     puiNew, 0, 0);
  Tcl_CreateCommand(i, "puiOpen",    puiOpen, 0, 0);
  Tcl_CreateCommand(i, "puiClose",   puiClose, 0, 0);

  Tcl_CreateCommand(i, "puiName",    puiName, 0, 0);
  Tcl_CreateCommand(i, "puiVersion", puiVersion, 0, 0);

  Tcl_CreateCommand(i, "puiFirst",   puiFirst, 0, 0);
  Tcl_CreateCommand(i, "puiLast",    puiLast, 0, 0);
  Tcl_CreateCommand(i, "puiPrev",    puiPrev, 0, 0);
  Tcl_CreateCommand(i, "puiNext",    puiNext, 0, 0);
  Tcl_CreateCommand(i, "puiList",    puiList, 0, 0);

  Tcl_CreateCommand(i, "puiRead",    puiRead, 0, 0);
  Tcl_CreateCommand(i, "puiWrite",   puiWrite, 0, 0);
  Tcl_CreateCommand(i, "puiDelete",  puiDelete, 0, 0);

  return TCL_OK;
}

static int puiNew(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  if (argc == 3) {
    PUI_FILE puidb;
    struct DBInfo info;
    int i;

    if ((i = dbentry(argv[2])) >= 0) {
      (*tclDBList[i]->NewDBInfo)(&info);
    } else {
      interp->result = "unknown database type.";
      return TCL_ERROR;
    }
    puidb = pui_Create(argv[1], &info);
    if (puidb == 0) {
      interp->result = "Couldn't open file";
      return TCL_ERROR;
    }

    sprintf(interp->result, "%lx", (unsigned long)puidb);
    return TCL_OK;
  } else {
    interp->result = "wrong # args: puiNew <filename> <databasename>";
    return TCL_ERROR;
  }
}

static int puiOpen(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  PUI_FILE puidb;

  if (argc == 2 || argc == 3) {
    if (argc == 2) {
      puidb = pui_Open(argv[1], "r"); /* DEFAULT TO READ-ONLY! */
    } else {
      puidb = pui_Open(argv[1], argv[2]);
    }
    if (puidb == 0) {
      interp->result = "Couldn't open file";
      return TCL_ERROR;
    }
    sprintf(interp->result, "%lx", (unsigned long)puidb);
    return TCL_OK;
  } else {
    interp->result = "wrong # args: puiOpen <filename> ?<mode>?";
    return TCL_ERROR;
  }
}

static int puiClose(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  if (argc == 2) {
    pui_Close((PUI_FILE)SAFESTRTOUL(argv[1], 16));
    return TCL_OK;
  } else {
    interp->result = "wrong # args: puiClose <db>";
    return TCL_ERROR;
  }
}

static int puiName(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  PUI_FILE puidb;
  if (argc == 2) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    Tcl_AppendElement(interp, pui_Name(puidb));
    return TCL_OK;
  } else {
    interp->result = "wrong # args: puiName <db>";
    return TCL_ERROR;
  }
}

static int puiVersion(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  PUI_FILE puidb;
  if (argc == 2) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    Tcl_AppendElement(interp, pui_Version(puidb));
    return TCL_OK;
  } else {
    interp->result = "wrong # args: puiVersion <db>";
    return TCL_ERROR;
  }
}

static int puiFirst(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  PUI_FILE puidb;
  unsigned long id;
  char s[16];

  if (argc == 2) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    id = pui_First(puidb); /* Skip deleted records? */
    sprintf(s, "%lx", id);
    Tcl_AppendElement(interp, s);
    return TCL_OK;
  } else {
    interp->result = "wrong # args: puiFirst <db>";
    return TCL_ERROR;
  }
}

static int puiLast(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  PUI_FILE puidb;
  unsigned long id;
  char s[16];

  if (argc == 2) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    id = pui_Last(puidb); /* Skip deleted records? */
    sprintf(s, "%lx", id);
    Tcl_AppendElement(interp, s);
    return TCL_OK;
  } else {
    interp->result = "wrong # args: puiLast <db>";
    return TCL_ERROR;
  }
}

static int puiPrev(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  PUI_FILE puidb;
  unsigned long id;
  char s[16];

  if (argc == 3) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    id = SAFESTRTOUL(argv[2], 16);
    if (pui_Exists(puidb, id)) {
      id = pui_Prev(puidb, id); /* Skip deleted records? */
      sprintf(s, "%lx", id);
      Tcl_AppendElement(interp, s);
      return TCL_OK;
    } else {
      interp->result = "bad id";
      return TCL_ERROR;
    }
  } else {
    interp->result = "wrong # args: puiPrev <db> <id>";
    return TCL_ERROR;
  }
}

static int puiNext(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  PUI_FILE puidb;
  unsigned long id;
  char s[16];

  if (argc == 3) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    id = SAFESTRTOUL(argv[2], 16);
    if (pui_Exists(puidb, id)) {
      id = pui_Next(puidb, id); /* Skip deleted records? */
      sprintf(s, "%lx", id);
      Tcl_AppendElement(interp, s);
      return TCL_OK;
    } else {
      interp->result = "bad id";
      return TCL_ERROR;
    }
  } else {
    interp->result = "wrong # args: puiNext <db> <id>";
    return TCL_ERROR;
  }
}

static int puiList(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  char s[16];
  PUI_FILE puidb;
  unsigned long id;
  int attr;
  int cat;

  if (argc == 2) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    id = pui_First(puidb);
    while(id) {
      pui_ReadHeader(puidb, id, &attr, &cat);
      if (!(attr & 0x80)) { /* If not deleted */
        sprintf(s, "%lx", (unsigned long)id);
        Tcl_AppendElement(interp, s);
      }
      id = pui_Next(puidb, id);
    }
    return TCL_OK;
  } else {
    interp->result = "wrong # args: puiList <db>";
    return TCL_ERROR;
  }
}

static int puiRead(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  char s[16];
  PUI_FILE puidb;
  unsigned long id;
  int attr, cat, size;
  char buffer[0xffff]; /* Big buffer to hold record */
  char *request;
  char *varname;
  int i;

  if (argc == 4) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    request = argv[2];
    varname = argv[3];

    /* USER REQUESTS DBINFO */
    if (strcmp(request, "DBInfo") == 0) {
      tclDBInfo_Read(interp, varname, puidb);
      return TCL_OK;
    }

    /* USER REQUESTS APPINFO */
    if (strcmp(request, "AppInfo") == 0) {
      if ((i = dbentry(pui_Name(puidb))) >= 0) {
        (*tclDBList[i]->AppInfoRead)(interp, varname, puidb);
      } else {
        interp->result = "unknown database type";
        return TCL_ERROR;
      }
      return TCL_OK;
    }

    /* USER REQUSTS RECORD */
    id = SAFESTRTOUL(request, 16);

    if (pui_Exists(puidb, id)) {
      pui_Read(puidb, id, &attr, &cat, &size, buffer);
      sprintf(s, "%d", attr);
      Tcl_SetVar2(interp, varname, "attribute", s, 0);
      sprintf(s, "%d", cat);
      Tcl_SetVar2(interp, varname, "category", s, 0);
      if ((i = dbentry(pui_Name(puidb))) >= 0) {
        (*tclDBList[i]->Read)(interp, varname, size, buffer);
      } else {
        interp->result = "unknown database type";
        return TCL_ERROR;
      }
      return TCL_OK;
    } else {
      interp->result = "bad id";
      return TCL_ERROR;
    }
  } else {
    interp->result = "wrong # args: puiRead <db> <id> <varname>\nReads record from database to array variable.\nDatabase information record can be read by setting <id> to \"DBInfo\"\n";
    return TCL_ERROR;
  }
}

static int puiWrite(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  char s[16];
  PUI_FILE puidb;
  unsigned long id;
  int attr, cat, size;
  char buffer[0xffff]; /* Big buffer to hold record */
  char *varname;
  char *request;
  char *value;
  int i;

  if (argc == 4) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    request = argv[2];
    varname = argv[3];

    /* USER REQUESTS DBINFO */
    if (strcmp(request, "DBInfo") == 0) {
      tclDBInfo_Write(interp, varname, puidb);
      return TCL_OK;
    }

    /* USER REQUESTS APPINFO */
    if (strcmp(request, "AppInfo") == 0) {
      if ((i = dbentry(pui_Name(puidb))) >= 0) {
        (*tclDBList[i]->AppInfoWrite)(interp, varname, puidb);
      } else {
        interp->result = "unknown database type";
        return TCL_ERROR;
      }
    }

    /* USER REQUESTS RECORD */
    id = SAFESTRTOUL(request, 16); /* If id == 0, a new record is created */
                                   /* user can also create a unique id. */

    value = Tcl_GetVar2(interp, varname, "attribute", 0);
    attr = (unsigned char)SAFEATOI(value);
    value = Tcl_GetVar2(interp, varname, "category", 0);
    cat = (unsigned char)SAFEATOI(value);
  
    size = 0;

    if ((i = dbentry(pui_Name(puidb))) >= 0) {
      (*tclDBList[i]->Write)(interp, varname, &size, buffer);
    } else {
      interp->result = "unknown database type";
      return TCL_ERROR;
    }

    attr |= 0x40; /* Mark record dirty */
    if (id && pui_Exists(puidb, id)) {
      pui_Write(puidb, id, attr, cat, size, buffer);
    } else {
      id = pui_Insert(puidb, (long)0, id, attr, cat, size, buffer);
    }

    sprintf(s, "%lx", (unsigned long)id); /* Return id, it may be new. */
    Tcl_AppendElement(interp, s);
    return TCL_OK;

  } else {
    interp->result = "wrong # args: puiWrite <db> <id> <varname>\nWrites record to database from data in array variable.\nDatabase information can be written by setting <id> to \"DBInfo\"\n";
    return TCL_ERROR;
  }
}

static int puiDelete(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
  PUI_FILE puidb;
  unsigned long id;
  int attr;
  int cat;
  int archive = 0;

  if (argc > 2 && argc < 5) {
    puidb = (PUI_FILE)SAFESTRTOUL(argv[1], 16);
    id = SAFESTRTOUL(argv[2], 16);
    if (argc == 4) {
      archive = atoi(argv[3]);
    }
    if (pui_Exists(puidb, id)) {
      pui_ReadHeader(puidb, id, &attr, &cat);
      attr |= 0x80 | 0x40;       /* Tag record as deleted & dirty */
      if (archive) attr |= 0x08; /* Tag "archived" if requested */
      pui_WriteHeader(puidb, id, attr, cat);
      return TCL_OK;
    } else {
      interp->result = "bad id";
      return TCL_ERROR;
    }
  } else {
    interp->result = "wrong # args: puiDelete <db> <id> ?<archive>?";
    return TCL_ERROR;
  }
}
